"""
Functions for reading from latest_forecast_values table
"""

import datetime as dt
import uuid
from typing import Dict, List, Optional

from pvsite_datamodel.sqlmodels import DatetimeIntervalSQL, LatestForecastValueSQL
from sqlalchemy.orm import Query, Session


def get_latest_forecast_values_by_site(
    session: Session, site_uuids: List[uuid.UUID], start_utc: Optional[dt.datetime] = None
) -> Dict[uuid.UUID, List[LatestForecastValueSQL]]:
    """
    Get the latest forecast values by input sites

    :param session: The sqlalchemy database session
    :param site_uuids: list of site_uuids for which to fetch latest forecast values
    :param start_utc: filters on forecast values target_time >= start_utc
    :return: dict containing {site_uuid1: List[LatestForecastValueSQL], site_uuid2: ...}
    """

    # start main query
    query: Query = session.query(LatestForecastValueSQL)

    if start_utc is not None:
        query = query.filter(LatestForecastValueSQL.target_time >= start_utc)

        # also filter on creation time, to speed up things
        created_utc_filter = start_utc - dt.timedelta(days=1)
        query = query.filter(LatestForecastValueSQL.created_utc >= created_utc_filter)

    output_dict: Dict[uuid.UUID, List[LatestForecastValueSQL]] = {}

    # Filter the query on the desired sites
    query = query.where(LatestForecastValueSQL.site_uuid.in_(site_uuids))

    # order by site, target time and created time desc
    query.join(DatetimeIntervalSQL)
    query.order_by(
        LatestForecastValueSQL.site_uuid,
        DatetimeIntervalSQL.start_utc,
        LatestForecastValueSQL.created_utc,
    )

    latest_forecast_values: List[LatestForecastValueSQL] = query.all()

    for site_uuid in site_uuids:
        site_latest_forecast_values: List[LatestForecastValueSQL] = [
            lfv for lfv in latest_forecast_values if lfv.site_uuid == site_uuid
        ]

        output_dict[site_uuid] = site_latest_forecast_values

    return output_dict



"""
Schemas and validation for dataframes used in the package
"""

import uuid

import pandera as pa
from pandera.dtypes import Bool, DateTime, Float, String
from pandera.typing import Series


class InputForecastValuesSchema(pa.SchemaModel):
    """Schema defining the makeup of the input forecast values dataframe"""

    target_datetime_utc: Series[DateTime] = pa.Field()
    forecast_kw: Series[Float] = pa.Field(ge=0)
    pv_uuid: Series[String] = pa.Field(le=32)

    @pa.check("pv_uuid")
    def pv_uuid_check(cls, series: Series[String]) -> Series[Bool]:
        """Check that pv_uuid column values are valid uuids"""
        try:
            uuid.UUID(series.str, version=4)
        except ValueError:
            return False
        return True



"""
Functions for wiritng datetime intervals
"""

import datetime as dt
import uuid

from pvsite_datamodel.sqlmodels import DatetimeIntervalSQL
from pvsite_datamodel.write.upsert import upsert
from pvsite_datamodel.write.utils import FORECAST_TIMESPAN, WrittenRow
from sqlalchemy import orm as sa_orm


def get_or_else_create_datetime_interval(
    session: sa_orm.Session, start_time: dt.datetime, end_time: dt.datetime | None = None
) -> tuple[DatetimeIntervalSQL, list[WrittenRow]]:
    """
    Gets a DatetimeInterval from the DB by start time if it exists, otherwise it creates a new entry

    :param session: The SQLAlchemy session used for performing db updates
    :param start_time: The start time of the datetime interval
    :param end_time: The end time of the datetime interval. Optional, defaults to the start_time
    + FORECAST_TIMESPAN
    :return tuple(DatetimeIntervalSQL, list[WrittenRow]): A tuple containing the existing
    or created DatetimeIntervalSQL object, and a list of WrittenRow objects dictating what was
    written to the DB
    """

    # End time defaults to the start time + FORECAST_TIMESPAN timedelta
    if end_time is None:
        end_time = start_time + FORECAST_TIMESPAN

    # Check if a datetime interval exists for the input times
    query = session.query(DatetimeIntervalSQL)
    query = query.filter(DatetimeIntervalSQL.start_utc == start_time)
    query = query.filter(DatetimeIntervalSQL.end_utc == end_time)
    existing_interval: DatetimeIntervalSQL = query.first()

    # If it does, fetch it's uuid
    if existing_interval is not None:
        return existing_interval, []

    # If it doesn't, create a new one
    else:
        datetime_interval: DatetimeIntervalSQL = DatetimeIntervalSQL(
            datetime_interval_uuid=uuid.uuid4(),
            start_utc=start_time,
            end_utc=end_time,
            created_utc=dt.datetime.now(tz=dt.timezone.utc),
        )
        written_rows = upsert(session, DatetimeIntervalSQL, [datetime_interval.__dict__])
        return datetime_interval, written_rows



"""
SQLAlchemy definition of the pvsite database schema
"""

import uuid
from datetime import datetime
from typing import List

import sqlalchemy as sa
from sqlalchemy import Column, DateTime
from sqlalchemy.dialects.postgresql import FLOAT, INTEGER, REAL, TIMESTAMP, UUID
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.schema import UniqueConstraint

Base = declarative_base()


class CreatedMixin:
    """Mixin to add created datetime to model"""

    created_utc = Column(DateTime(timezone=True), default=lambda: datetime.utcnow())


class SiteSQL(Base, CreatedMixin):
    """
    Class representing the sites table.

    Each site row specifies a single panel or cluster of panels
    found on a residential house or commercial building. Their
    data is provided by a client.

    *Approximate size: *
    4 clients * ~1000 sites each = ~4000 rows
    """

    __tablename__ = "sites"

    site_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    client_uuid = sa.Column(
        UUID(as_uuid=True), sa.ForeignKey("clients.client_uuid"), nullable=False, default=uuid.uuid4
    )
    client_site_id = sa.Column(INTEGER)
    client_site_name = sa.Column(sa.String(255))
    region = sa.Column(sa.String(255))
    dno = sa.Column(sa.String(255))
    gsp = sa.Column(sa.String(255))
    orientation = sa.Column(REAL)
    tilt = sa.Column(REAL)
    latitude = sa.Column(FLOAT, nullable=False)
    longitude = sa.Column(FLOAT, nullable=False)
    capacity_kw = sa.Column(REAL, nullable=False)
    updated_utc = sa.Column(TIMESTAMP, nullable=False, default=lambda: datetime.utcnow())
    ml_id = sa.Column(sa.Integer, autoincrement=True, nullable=False, unique=True)

    __table_args__ = (UniqueConstraint("client_site_id", client_uuid, name="idx_client"),)

    forecasts: List["ForecastSQL"] = relationship("ForecastSQL")
    latest_forecast_values: List["LatestForecastValueSQL"] = relationship("LatestForecastValueSQL")
    generation: List["GenerationSQL"] = relationship("GenerationSQL")


class GenerationSQL(Base, CreatedMixin):
    """
    Class representing the generation table.

    Each generation row specifies a generated power output over a
    given time range for a site.

    *Approximate size: *
    Generation populated every 5 minutes per site * 4000 sites = ~1,125,000 rows per day
    """

    __tablename__ = "generation"

    generation_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    site_uuid = sa.Column(
        UUID(as_uuid=True), sa.ForeignKey("sites.site_uuid"), nullable=False, default=uuid.uuid4
    )
    power_kw = sa.Column(REAL, nullable=False)
    datetime_interval_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("datetime_intervals.datetime_interval_uuid"),
        nullable=False,
        default=uuid.uuid4,
    )


class ForecastSQL(Base, CreatedMixin):
    """
    Class representing the forecasts table.

    Each forecast row refers to a sequence of predicted solar generation values
    over a set of target times for a site.

    *Approximate size: *
    One forecast per site every 5 minutes = ~1,125,000 rows per day
    """

    __tablename__ = "forecasts"

    forecast_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    site_uuid = sa.Column(
        UUID(as_uuid=True), sa.ForeignKey("sites.site_uuid"), nullable=False, default=uuid.uuid4
    )

    forecast_version = sa.Column(sa.String(32), nullable=False)

    # one (forecasts) to many (forecast_values)
    forecast_values: List["ForecastValueSQL"] = relationship("ForecastValueSQL")


class ForecastValueSQL(Base, CreatedMixin):
    """
    Class representing the forecast_values table.

    Each forecast_value row is a prediction for the power output
    of a site over a target datetime interval. Many predictions
    are made for each site at each target interval.

    *Approximate size: *
    One forecast value every 5 minutes per site per forecast.
    Each forecast's prediction sequence covers 24 hours of target
    intervals = ~324,000,000 rows per day
    """

    __tablename__ = "forecast_values"

    forecast_value_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    datetime_interval_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("datetime_intervals.datetime_interval_uuid"),
        nullable=False,
        default=uuid.uuid4,
    )
    forecast_generation_kw = sa.Column(REAL, nullable=False)

    forecast_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("forecasts.forecast_uuid"),
        nullable=False,
        default=uuid.uuid4,
    )


class LatestForecastValueSQL(Base, CreatedMixin):
    """
    Class representing the latest_forecast_values table.

    Each latest_forecast_value row is a prediction for the power output
    of a site over a target datetime interval. Only the most recent
    prediction for each target time interval is stored in this table
    per site.

    *Approximate size: *
    One forecast value every 5 minutes per site per forecast
    sequence = ~1,125,000 rows per day
    """

    __tablename__ = "latest_forecast_values"

    latest_forecast_value_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    datetime_interval_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("datetime_intervals.datetime_interval_uuid"),
        nullable=False,
        default=uuid.uuid4,
    )
    forecast_generation_kw = sa.Column(REAL, nullable=False)

    forecast_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, nullable=False)
    site_uuid = sa.Column(
        UUID(as_uuid=True), sa.ForeignKey("sites.site_uuid"), nullable=False, default=uuid.uuid4
    )
    forecast_version = sa.Column(sa.String(32), nullable=False)

    latest_forecast: SiteSQL = relationship("SiteSQL", back_populates="latest_forecast_values")


class ClientSQL(Base, CreatedMixin):
    """
    Class representing the clients table.

    Each client row defines a provider of site data

    *Approximate size: *
    One row per client = ~4 rows
    """

    __tablename__ = "clients"

    client_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    client_name = sa.Column(sa.String(255), nullable=False)

    sites: List[SiteSQL] = relationship("SiteSQL")


class DatetimeIntervalSQL(Base, CreatedMixin):
    """
    Class representing the datetime_intervals table.

    Each datetime_interval row defines a timespan between a start and end time

    *Approximate size: *
    One interval every 5 minutes per day = ~288 rows per day
    """

    __tablename__ = "datetime_intervals"

    datetime_interval_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    start_utc = sa.Column(TIMESTAMP, nullable=False)
    end_utc = sa.Column(TIMESTAMP, nullable=False)

    generation: List[GenerationSQL] = relationship("GenerationSQL")
    forecast_values: List[ForecastValueSQL] = relationship("ForecastValueSQL")
    latest_forecast_values: List[LatestForecastValueSQL] = relationship("LatestForecastValueSQL")


class StatusSQL(Base, CreatedMixin):
    """
    Class representing the status table:

    Each status row defines a message reporting on the status of the
    services within the nowcasting domain

    *Approximate size: *
    ~1 row per day
    """

    __tablename__ = "status"

    status_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    status = sa.Column(sa.String(255))
    message = sa.Column(sa.String(255))



"""
Generic DB rows writer
"""

import sqlalchemy as sa
from pvsite_datamodel.sqlmodels import Base
from pvsite_datamodel.write.utils import WrittenRow
from sqlalchemy import orm as sa_orm
from sqlalchemy.dialects import postgresql as sa_psql


def upsert(session: sa_orm.Session, table: Base, rows: list[dict]) -> list[WrittenRow]:
    """
    Upsert rows into table

    This functions checks the primary keys, and if present, updates the row.
    :param session: sqlalchemy Session
    :param table: the table
    :param rows: the rows we are going to update
    :return list[WrittenRow]: A list of WrittenRow objects containing the table and primary
    key values
    that have been written
    """

    # Input type validation in case user passes in a dict, not a list of dicts
    if type(rows) != list:
        if type(rows) == dict:
            rows = [rows]
        else:
            raise TypeError("input rows must be a list of dict objects")

    stmt = sa_psql.insert(table.__table__)
    primary_key_names = [key.name for key in sa.inspect(table.__table__).primary_key]
    update_dict = {c.name: c for c in stmt.excluded if not c.primary_key}

    if not update_dict:
        raise ValueError("insert_or_update resulted in an empty update_dict")

    stmt = stmt.on_conflict_do_update(index_elements=primary_key_names, set_=update_dict)
    session.execute(stmt, rows)
    session.commit()

    return [WrittenRow(table=table, pk_value=row[primary_key_names[0]]) for row in rows]



"""
Functions for writing to pvsite db
"""

import datetime as dt
import logging
import uuid

import dateutil.parser as dp
import numpy.typing as npt
import pandas as pd
import sqlalchemy.orm as sa_orm
from pvsite_datamodel.read.site import get_site_by_uuid
from pvsite_datamodel.sqlmodels import ForecastSQL, ForecastValueSQL
from pvsite_datamodel.write.datetime_intervals import get_or_else_create_datetime_interval
from pvsite_datamodel.write.upsert import upsert
from pvsite_datamodel.write.utils import WrittenRow


def insert_forecast_values(
    session: sa_orm.Session, df_forecast_values: pd.DataFrame, forecast_version: str = "0.0.0"
) -> list[WrittenRow]:
    """
    Inserts a dataframe of forecast values into the database.

    :param session: sqlalchemy session for interacting with the database
    :param df_forecast_values: pandas dataframe with columns
    ["target_datetime_utc", "forecast_kw", "pv_uuid"]
    :param forecast_version: the version of the model used to create the forecast
    :return list[WrittenRow]: list of added rows to DB
    """

    # Track rows added to DB
    written_rows: list[WrittenRow] = []

    # Loop over all the unique sites that have got forecast values
    sites: npt.ndarray[uuid.UUID] = df_forecast_values["pv_uuid"].unique()
    sql_forecast_values = []
    for site_uuid in sites:

        # Check whether the site id exits in the table, otherwise return an error
        get_site_by_uuid(session=session, site_uuid=site_uuid)

        # Create a forcast (sequence) for the site, and write it to db
        forecast: ForecastSQL = ForecastSQL(
            forecast_uuid=uuid.uuid4(),
            site_uuid=site_uuid,
            created_utc=dt.datetime.now(tz=dt.timezone.utc),
            forecast_version=forecast_version,
        )
        newly_written_rows = upsert(session, ForecastSQL, forecast.__dict__)
        written_rows.extend(newly_written_rows)

        # Get all dataframe forecast value entries for current site_uuid
        df_site: pd.DataFrame = df_forecast_values.loc[df_forecast_values["pv_uuid"] == site_uuid]

        # Filter the forecasted values by target_time
        target_times: npt.ndarray[dt.datetime] = df_site["target_datetime_utc"].unique()

        # Print a warning if there are duplicate target_times for this site's forecast
        if len(target_times) != len(df_site):
            logging.warning(
                f"duplicate target times exist in forecast {forecast.forecast_uuid} "
                f"for site {site_uuid}"
            )

        # For each target time:
        for target_time in target_times:
            # Convert the target_time string to a datetime object
            target_time_as_datetime: dt.datetime = dp.isoparse(target_time)

            datetime_interval, newly_added_rows = get_or_else_create_datetime_interval(
                session=session, start_time=target_time_as_datetime
            )
            written_rows.extend(newly_added_rows)

            # For each entry with this targettime (there should only be one)
            df_target_entries: pd.DataFrame = df_site.loc[
                df_site["target_datetime_utc"] == target_time
            ]

            # Create a ForecastValueSQL object for each forecast value, and surface as dict
            # TODO there might be quicker ways to do this, like go from pandas straight to a dict
            sql_forecast_value = ForecastValueSQL(
                forecast_uuid=forecast.forecast_uuid,
                forecast_value_uuid=uuid.uuid4(),
                datetime_interval_uuid=datetime_interval.datetime_interval_uuid,
                created_utc=dt.datetime.now(tz=dt.timezone.utc),
                forecast_generation_kw=df_target_entries.iloc[0]["forecast_kw"],
            ).__dict__
            sql_forecast_values.append(sql_forecast_value)

    # Save it to the db
    newly_added_rows = upsert(session, ForecastValueSQL, sql_forecast_values)
    written_rows.extend(newly_added_rows)

    return written_rows



"""Test write functions"""

import pandas as pd
import pytest

from pvsite_datamodel.sqlmodels import (
    SiteSQL,
    DatetimeIntervalSQL,
    ForecastSQL,
    ForecastValueSQL,
    GenerationSQL,
)

from pvsite_datamodel import DatabaseConnection
from pvsite_datamodel.write.datetime_intervals import get_or_else_create_datetime_interval
from pvsite_datamodel.write import insert_forecast_values
from pvsite_datamodel.write import insert_generation_values


class TestDatabaseConnection:
    """Tests for the DatabaseConnection class"""

    def test_connection(self, engine, sites):
        dbcon = DatabaseConnection(engine.url, echo=False)
        with dbcon.get_session() as session:
            session.query(SiteSQL).first()


class TestGetOrElseCreateDatetimeInterval:
    """Tests for the get_or_else_create_datetime_interval function"""

    def test_writes_interval_when_not_exists(self, db_session, test_time):
        # Call function with test_time
        datetime_interval, written_rows = get_or_else_create_datetime_interval(
            session=db_session, start_time=test_time
        )
        assert len(written_rows) > 0

        # Check data has been written and exists in table
        query_result = (
            db_session.query(DatetimeIntervalSQL)
            .filter(
                DatetimeIntervalSQL.datetime_interval_uuid
                == datetime_interval.datetime_interval_uuid
            )
            .first()
        )
        assert query_result is not None

    def test_gets_existing_interval_when_exists(self, db_session, test_time):
        _, written_rows = get_or_else_create_datetime_interval(
            session=db_session, start_time=test_time
        )
        _, written_rows_2 = get_or_else_create_datetime_interval(
            session=db_session, start_time=test_time
        )
        assert len(written_rows) > 0
        assert len(written_rows_2) == 0


class TestInsertForecastValues:
    """Tests for the insert_forecast_values function"""

    def test_inserts_values_for_existing_site(self, db_session, forecast_valid_site):
        df = pd.DataFrame(forecast_valid_site)
        written_rows = insert_forecast_values(session=db_session, df_forecast_values=df)

        assert len(written_rows) == 21
        # 10 datetime intervals, 10 forecast values, 1 forecast

        # Check data has been written and exists in table
        written_forecastvalues = db_session.query(ForecastValueSQL).all()
        assert len(written_forecastvalues) == 10
        written_datetimeintervals = db_session.query(DatetimeIntervalSQL).all()
        assert len(written_datetimeintervals) == 10
        written_forecasts = db_session.query(ForecastSQL).all()
        assert len(written_forecasts) == 1

    def test_inserts_values_for_existing_site_and_existing_datetime_intervals(
            self, db_session, forecast_valid_site
    ):
        """
        Tests inserts values successfully without creating redundant datetime intervals
        """

        # Create DataFrame and write to DB
        df = pd.DataFrame(forecast_valid_site)
        # write once and again,
        _ = insert_forecast_values(session=db_session, df_forecast_values=df)
        written_rows = insert_forecast_values(session=db_session, df_forecast_values=df)
        assert len(written_rows) == 11  # 10 forecast values, 1 forecast

        # Check correct data has been written and exists in table
        written_forecastvalues = db_session.query(ForecastValueSQL).all()
        assert len(written_forecastvalues) == 20  # 10 more since previous test

        written_datetimeintervals = db_session.query(DatetimeIntervalSQL).all()
        assert len(written_datetimeintervals) == 10  # Unchanged from previous test

        written_forecasts = db_session.query(ForecastSQL).all()
        assert len(written_forecasts) == 2  # 1 more since previous test

    def test_errors_on_invalid_site(self, db_session, forecast_invalid_site):
        """Tests function errors when incoming pv_uuid does not exist in sites table"""

        df = pd.DataFrame(forecast_invalid_site)
        with pytest.raises(KeyError):
            written_rows = insert_forecast_values(session=db_session, df_forecast_values=df)
            assert len(written_rows), 0


class TestInsertGenerationValues:
    """Tests for the insert_generation_values function"""

    def test_inserts_generation_for_existing_site(self, db_session, generation_valid_site):
        """Tests inserts values successfully"""

        df = pd.DataFrame(generation_valid_site)
        written_rows = insert_generation_values(session=db_session, generation_values_df=df)

        assert len(written_rows) == 20
        # 10 datetime intervals, 10 generation values

        # Check data has been written and exists in table
        written_forecastvalues = db_session.query(GenerationSQL).all()
        assert len(written_forecastvalues) == 10
        written_datetimeintervals = db_session.query(DatetimeIntervalSQL).all()
        assert len(written_datetimeintervals) == 10



"""
Cross-package functions
"""

from datetime import datetime
from typing import Optional

from pvsite_datamodel.sqlmodels import DatetimeIntervalSQL
from sqlalchemy.orm import Query


def filter_query_by_datetime_interval(
    query: Query, start_utc: Optional[datetime] = None, end_utc: Optional[datetime] = None
) -> Query:
    """
    Applies a filter on the input query according to the Datetime Interval table.

    Ensure DatetimeIntervalSQL has been joined to query already
    :param query: The sqlalchemy query
    :param start_utc: search filters >= on 'start_utc'. Can be None
    :param end_utc: search filters < on 'end_utc'. Can be None
    :return: The sqlalchemy query
    """

    # filter on start time
    if start_utc is not None:
        query = query.filter(DatetimeIntervalSQL.start_utc >= start_utc)

    # filter on end time
    if end_utc is not None:
        query = query.filter(DatetimeIntervalSQL.end_utc < end_utc)

    return query
